package server.ai;

import java.util.function.Predicate;

import common.ai.AIRangedAttack;
import common.ai.EntityAIAttackOnCollide;
import common.ai.EntityAIAvoidEntity;
import common.ai.EntityAIHurtByTarget;
import common.ai.EntityAILookAtTalkingPlayer;
import common.ai.EntityAINagPlayer;
import common.ai.EntityAINearestAttackableTarget;
import common.ai.EntityAINpcInteract;
import common.ai.EntityAINpcMate;
import common.ai.EntityAIOpenDoor;
import common.ai.EntityAIPlay;
import common.ai.EntityAISwimming;
import common.ai.EntityAIWander;
import common.ai.EntityAIWatchClosest;
import common.ai.EntityAIWatchClosest2;
import common.ai.IEntityNPCNode;
import common.entity.npc.Alignment;
import common.entity.npc.EntityNPC;
import common.entity.types.EntityLiving;
import common.init.Items;
import common.item.ItemGunBase;
import common.item.ItemStack;
import common.pathfinding.PathNavigateGround;

public class EntityNPCNode<T extends EntityNPC> extends EntityLivingNode<T> implements IEntityNPCNode {
    protected final EntityAIAttackOnCollide aiMelee;
    protected final AIRangedAttack aiRanged;
    private boolean fleeing;
	
	public EntityNPCNode(T entity) {
		super(entity);
		this.aiMelee = new EntityAIAttackOnCollide(entity, EntityLiving.class, 1.0D, true, true) {
	    	public boolean shouldExecute()
	        {
	        	return !EntityNPCNode.this.fleeing && super.shouldExecute();
	        }
	    	
	        public boolean continueExecuting()
	        {
	        	return !EntityNPCNode.this.fleeing && super.shouldExecute();
	        }
	    };
	    this.aiRanged = new AIRangedAttack(entity, 1.0D, entity.getAttackSpeed(), entity.getAttackSpeed() * 3, 15.0F) {
//	    	public void startExecuting()
//	        {
//	            super.startExecuting();
//	            EntityNPC.this.setUsingItem(true);
//	        }
//	    	
//	    	public void resetTask()
//	        {
//	            super.resetTask();
//	            EntityNPC.this.setUsingItem(false);
//	        }
	    	
	        public boolean shouldExecute()
	        {
	        	return !EntityNPCNode.this.fleeing && super.shouldExecute();
	        }
	    	
	        public boolean continueExecuting()
	        {
	        	return !EntityNPCNode.this.fleeing && super.shouldExecute();
	        }
	    };
	}
	
	protected void registerTasks() {
        if(this.getNavigator() instanceof PathNavigateGround) {
	        ((PathNavigateGround)this.getNavigator()).setBreakDoors(true);
	        ((PathNavigateGround)this.getNavigator()).setAvoidsWater(true);
        }
        this.tasks.addTask(0, new EntityAISwimming(this.entity));
//        this.tasks.addTask(1, new EntityAIAttackOnCollide(this, EntityNPC.class, 0.6D, true, true));
//        this.tasks.addTask(1, new EntityAIAttackOnCollide(this, EntityLivingBase.class, 0.6D, true, true));
        this.tasks.addTask(1, new EntityAINpcMate(this.entity));
        this.tasks.addTask(2, new EntityAIAvoidEntity(this.entity, EntityLiving.class, new Predicate<EntityLiving>() {
			public boolean test(EntityLiving entity) {
				return entity != EntityNPCNode.this.entity && EntityNPCNode.this.shouldFlee(entity);
			}
		}, 8.0F, 1.1D, 1.1D) {
        	{
        		this.setMutexBits(0);
        	}
        	
        	public void startExecuting() {
                EntityNPCNode.this.setAttackTarget(null);
                EntityNPCNode.this.fleeing = true;
            }
        	
        	public boolean continueExecuting() {
                return false;
            }

            public void updateTask()
            {
            }
        	
//        	public void resetTask() {
//                super.resetTask();
//                EntityNPC.this.isFleeing = false;
//            }
        });
        this.tasks.addTask(3, new EntityAIAvoidEntity(this.entity, EntityLiving.class, new Predicate<EntityLiving>() {
			public boolean test(EntityLiving entity) {
				return entity != EntityNPCNode.this.entity && EntityNPCNode.this.shouldFlee(entity);
			}
		}, 8.0F, 1.1D, 1.1D) {
        	public void startExecuting() {
        		super.startExecuting();
        		EntityNPCNode.this.setAttackTarget(null);
        		EntityNPCNode.this.fleeing = true;
            }
        	
        	public void resetTask() {
                super.resetTask();
                EntityNPCNode.this.fleeing = false;
            }
        });
//        this.tasks.addTask(2, new EntityAITempt(this, 1.0D, Items.golden_apple, false));
        this.tasks.addTask(4, new EntityAIOpenDoor(this.entity, true));
        this.tasks.addTask(5, new EntityAINagPlayer(this.entity));
        this.tasks.addTask(5, new EntityAILookAtTalkingPlayer(this.entity));
        this.tasks.addTask(6, new EntityAIWatchClosest2(this.entity, null, 3.0F, 1.0F) {
        	private int sneakTime;
        	
        	public void updateTask()
            {
                super.updateTask();
                boolean flag = this.closestEntity.isPlayer() && this.closestEntity.isSneaking();
                if(this.sneakTime > 0) {
                	if(--this.sneakTime == 0) {
                		EntityNPCNode.this.entity.setSneaking(false);
                	}
                }
                else if(EntityNPCNode.this.getAttackTarget() == null && EntityNPCNode.this.rand.chance(flag ? 5 : 200)) {
                	EntityNPCNode.this.entity.setSneaking(true);
                	this.sneakTime = EntityNPCNode.this.rand.range(60, flag ? 160 : 120);
                }
                else if(EntityNPCNode.this.getAttackTarget() != null) {
                	EntityNPCNode.this.entity.setSneaking(false);
                	this.sneakTime = 0;
                }
            }
        	
        	public void resetTask()
            {
                super.resetTask();
                EntityNPCNode.this.entity.setSneaking(false);
            }
        });
        this.tasks.addTask(7, new EntityAINpcInteract(this.entity));
        this.tasks.addTask(8, new EntityAIWander(this.entity, 1.0D));
        this.tasks.addTask(8, new EntityAIPlay(this.entity, 1.1D));
        this.tasks.addTask(9, new EntityAIWatchClosest(this.entity, EntityLiving.class, 8.0F));
        this.targets.addTask(1, new EntityAIHurtByTarget(this.entity, false, /* EntityNPC.class, */ EntityLiving.class) {
            protected boolean isSuitableTarget(EntityLiving entity) {
            	if(entity != null && entity != EntityNPCNode.this.entity && EntityNPCNode.this.shouldFlee(entity)) {
            		EntityNPCNode.this.setAttackTarget(null);
            		EntityNPCNode.this.fleeing = true;
            		return false;
            	}
				return entity != null && entity != EntityNPCNode.this.entity && (!(entity.isPlayer()) || EntityNPCNode.this.rand.chance(entity.getAttackedBy() == EntityNPCNode.this.entity ? 2 : 4))
						&& EntityNPCNode.this.canCounter(entity) && super.isSuitableTarget(entity);
            }
        });
        this.targets.addTask(2, new EntityAINearestAttackableTarget(this.entity, EntityLiving.class, 10, true, false, new Predicate<EntityLiving>() {
			public boolean test(EntityLiving entity) {
            	if(entity != null && entity != EntityNPCNode.this.entity && EntityNPCNode.this.shouldFlee(entity)) {
            		EntityNPCNode.this.setAttackTarget(null);
            		EntityNPCNode.this.fleeing = true;
            		return false;
            	}
				return entity != null && entity != EntityNPCNode.this.entity && !EntityNPCNode.this.fleeing && EntityNPCNode.this.entity.canAmbush(entity) && EntityNPCNode.this.canAttack(entity);
			}
		}));
//        this.setCanPickUpLoot(true);
        
        this.setCombatTask();
	}

    public final boolean canInfight(Alignment align) {
    	Alignment alignment = this.entity.getAlignmentCached();
    	return alignment.chaotic || (!alignment.lawful && 
    			((alignment.good && align.evil) || (alignment.evil && align.good)));
    }
    
    public final boolean canMurder(Alignment align) {
    	Alignment alignment = this.entity.getAlignmentCached();
    	return alignment.chaotic && ((alignment.good && align.evil) || (alignment.evil && align.good));
    }
    
    public boolean canCounter(EntityLiving entity) {
		return !(entity instanceof EntityNPC) || 
			(this.entity.getClass() == entity.getClass() ? this.canInfight(((EntityNPC)entity).getAlignmentCached()) : 
				!this.entity.isPeaceful(((EntityNPC)entity).getClass()));
    }
    
    public boolean canAttack(EntityLiving entity) {
		return entity instanceof EntityNPC && (this.entity.getClass() == entity.getClass() ? this.canMurder(((EntityNPC)entity).getAlignmentCached()) : 
				this.entity.isAggressive(((EntityNPC)entity).getClass()));
    }
    
    public boolean shouldFlee(EntityLiving entity) {
    	return (this.entity.getHealth() <= (this.entity.getMaxHealth() / 4) || !this.canCounter(entity)) && 
    			((entity instanceof EntityNPC && ((EntityNPC)entity).getNode().canAttack(this.entity) || 
    			(entity == this.entity.getAttackedBy() && ((EntityNPC)entity).getNode().canCounter(this.entity))));
    }
    
    public boolean isRangedWeapon(ItemStack stack) {
    	return stack != null && (stack.getItem() == Items.bow || stack.getItem() instanceof ItemGunBase || 
    			stack.getItem() == Items.snowball || stack.getItem() == Items.potion);
    }
    
//    public boolean isRangedWeapon(ItemStack stack) { // EntityGargoyle
//    	return false;
//    }
    
	public void updateLeashedState() {
		if(!this.player)
			super.updateLeashedState();
	}
	
    public void sendDeathMessage() {
		this.sendDeathMessage(this.player, true);
	}
    
    public int getMaxFallHeight()
    {
        return this.player ? 3 : super.getMaxFallHeight();
    }
    
    public void setCombatTask()
    {
    	if(!this.player) {
            this.tasks.removeTask(this.aiMelee);
            this.tasks.removeTask(this.aiRanged);
            ItemStack itemstack = this.entity.getHeldItem();

            if (this.isRangedWeapon(itemstack))
            {
                this.tasks.addTask(3, this.aiRanged);
            }
            else
            {
                this.tasks.addTask(3, this.aiMelee);
            }
    	}
    }

	public boolean isFleeing() {
		return this.fleeing;
	}
}
